#!/usr/bin/env python3
#
# Sniffles2
# A fast structural variant caller for long-read sequencing data
#
# Created: 27.08.2021
# Author:  Moritz Smolka
# Contact: moritz.g.smolka@gmail.com
#

import statistics
import math
import sys
import time

class Sniffles2Exit(Exception):
    pass

def stdev(nums):
    nums=list(nums)
    return statistics.stdev(nums) if len(nums) > 1 else 0

def median(nums):
    return int(statistics.median(nums))

def median_or_mode(nums):
    nums=list(nums)
    top=most_common(nums)
    if len(top)>1 and (top[0][0]-top[1][0]<2):
    #if len(nums)>3:
        return median_noavg(nums)
    else:
        return median_modes(nums)

def median_noavg(nums):
    nums=sorted(list(nums))
    mid=int(len(nums)/2)
    return nums[mid]

def median_modes(nums):
    max_count=0
    counts={}
    for n in nums:
        if not n in counts:
            counts[n]=1
        else:
            counts[n]+=1
        max_count=max(max_count,counts[n])
    return median_noavg(k for k,n in counts.items() if max_count-n<3 )

def mean(nums):
    nums=list(nums)
    return sum(nums)/len(nums)

def mean_or_none(nums):
    nums=list(nums)
    if len(nums)==0:
        return None
    else:
        return sum(nums)/len(nums)

def mean_or_none_round(nums):
    r=mean_or_none(nums)
    if r==None:
        return r
    else:
        return round(r)

def trim(nums,pct=25):
    nums=sorted(list(nums))
    trim_n=int(len(nums)/float(100.0)*pct)
    if trim_n > 0:
        return nums[trim_n:-trim_n]
    else:
        return nums

def most_common(nums):
    counts={}
    for n in nums:
        if not n in counts:
            counts[n]=1
        else:
            counts[n]+=1
    return sorted(((counts[n],n) for n in counts), reverse=True)

def most_common_top(nums):
    result=most_common(nums)
    return sorted(item for count,item in result if count==result[0][0])[0]

def error(msg):
    sys.stderr.write("Sniffles2 Error: "+msg+"\n")
    sys.stderr.flush()

def fatal_error(msg):
    error(msg+" (Fatal error, exiting.)")
    exit(1)

def fatal_error_main(msg):
    error(msg+" (Fatal error, exiting.)")
    raise Sniffles2Exit

def load_tandem_repeats(filename,padding):
    contigs_tr={}
    unsorted=False
    with open(filename,"r") as handle:
        for line in handle:
            parts=line.split("\t")
            if len(parts) >= 3:
                contig,start,end=parts[:3]
                start=int(start)
                end=int(end)
                if not contig in contigs_tr:
                    contigs_tr[contig]=[]
                if len(contigs_tr[contig])>0:
                    last_start,last_end=contigs_tr[contig][-1]
                    if start < last_start:
                        unsorted=True
                contigs_tr[contig].append((max(0,int(start)-padding),int(end)+padding))

    if unsorted:
        print("Info: The tandem repeat annotations file was not sorted. Sorting it in-memory after loading... (please sort the .bed file once before to save time when running multiple samples)")
        sort_start=time.time()
        for contig in contigs_tr:
            contigs_tr[contig].sort()
        print(f"Info: Optional sorting of input tandem repeat annotations took {time.time()-sort_start:.2f}s.")

    return contigs_tr

center=median_modes



#!/usr/bin/env python3
#
# Sniffles2
# A fast structural variant caller for long-read sequencing data
#
# Created: 15.08.2021
# Author:  Moritz Smolka
# Contact: moritz.g.smolka@gmail.com
#

from dataclasses import dataclass

from sniffles import util

TYPES=["INS","DEL","DUP","INV","BND"]

@dataclass
class SVCallBNDInfo:
    mate_contig: str
    mate_ref_start: int
    is_first: bool
    is_reverse:bool

@dataclass
class SVCallPostprocessingInfo:
    cluster: list

@dataclass
class SVCall:
    contig: str
    pos: int
    id: str
    ref: str
    alt: str
    qual: int
    filter: str
    info: dict

    svtype: str
    svlen: int
    end: int
    genotypes: dict

    precise: bool
    support: int
    rnames: list

    qc: bool
    nm: float
    postprocess: SVCallPostprocessingInfo

    fwd: int = None
    rev: int = None

    coverage_upstream: int = None
    coverage_downstream: int = None
    coverage_start: int = None
    coverage_center: int = None
    coverage_end: int = None

    def set_info(self,k,v):
        self.info[k]=v

    def get_info(self,k):
        return self.info[k] if k in self.info else None

    def has_info(self,k):
        return k in self.info

    def finalize(self):
        self.postprocess=None

@dataclass
class SVGroup:
    candidates: list
    pos_mean: float
    len_mean: float
    included_samples: set
    coverages_nonincluded: dict

def calculate_bounds(svtype,ref_start_mode,svlen_mode):
    if svtype=="INS":
        svstart=ref_start_mode
        svend=ref_start_mode
    elif svtype=="DEL":
        svstart=ref_start_mode+svlen_mode
        svend=ref_start_mode
    else:
        svstart=ref_start_mode
        svend=svstart+abs(svlen_mode)
    return svstart,svend

def call_from(cluster,config,keep_qc_fails,task):
    leads=cluster.leads

    svtype=cluster.svtype
    stdev_pos,stdev_len=None,None
    qc=True

    svlen=util.center(v.svlen for v in leads)

    if abs(svlen) < config.minsvlen:
        return

    #Count inline events only once per read, but split events as individual alignments, as in coverage calculation
    #inline_qnames=set(k.read_qname for k in leads if k.source=="INLINE")
    #support=len(inline_qnames)+sum(1 for k in leads if k.source!="INLINE")
    if svtype=="INS" and svlen>=config.long_ins_length:
        support_long_set=set(lead.read_qname for lead in cluster.leads_long)
        support=len(set(k.read_qname for k in leads) | support_long_set)
        support_long=len(support_long_set)
    else:
        support=len(set(k.read_qname for k in leads))
        support_long=0
    ref_start=util.center(v.ref_start for v in leads)
    stdev_pos=util.stdev(util.trim((v.ref_start for v in leads)))
    if svtype!="BND":
        stdev_len=util.stdev(util.trim((v.svlen for v in leads)))
        precise=(stdev_pos+stdev_len < config.precise)
    else:
        precise=(stdev_pos < config.precise)
    svstart,svend=calculate_bounds(svtype,ref_start,svlen)
    qual=int(util.mean(v.mapq for v in leads))

    support_fwd=sum(lead.strand == "+" for lead in leads)
    support_rev=len(leads) - support_fwd

    if config.qc_strand and (support_fwd==0 or support_rev==0):
        qc=False

    if config.qc_nm:
        nm_mean=util.mean(v.nm for v in leads)
        if nm_mean > config.qc_nm_max:
            qc=False
    else:
        nm_mean=-1

    if not keep_qc_fails and not qc:
        return

    svpi=SVCallPostprocessingInfo(cluster=cluster)

    if config.output_rnames or config.snf!=None:
        rnames=list(set(k.read_qname for k in leads))
    else:
        rnames=None

    svcall=SVCall(contig=cluster.contig,
                  pos=svstart,
                  id=f"{svtype}.{task.sv_id:X}S{task.id:X}",
                  ref="N",
                  alt=f"<{svtype}>",
                  qual=qual,
                  filter="PASS",
                  info=dict(),
                  svtype=svtype,
                  svlen=svlen,
                  end=svend,
                  genotypes=dict(),
                  precise=precise,
                  support=support,
                  rnames=rnames,
                  postprocess=svpi,
                  qc=qc,
                  nm=nm_mean,
                  fwd=support_fwd,
                  rev=support_rev)

    if svtype=="BND":
        resolve_bnd(svcall,cluster,config)
    elif svtype=="INS":
        svcall.set_info("SUPPORT_LONG",support_long)

    if stdev_pos!=None:
        svcall.set_info("STDEV_POS",stdev_pos)
    if stdev_len!=None:
        svcall.set_info("STDEV_LEN",stdev_len)

    task.sv_id+=1

    yield svcall

def merge_inner_bounds(leads,config):
    svlength_binsize=25
    read_svlengths={}
    read_starts={}
    for ld in leads:
        if not ld.read_id in read_svlengths:
            read_svlengths[ld.read_id]=0
            read_starts[ld.read_id]=ld.ref_start
        read_svlengths[ld.read_id]+=ld.svlen
        read_starts[ld.read_id]=min(ld.ref_start,read_starts[ld.read_id])
    svlen=util.center(v for k,v in read_svlengths.items())
    pos=util.center(v for k,v in read_starts.items())

    svlengths_reads={}
    for read_id, length in read_svlengths.items():
        bin=int(length/svlength_binsize)*svlength_binsize
        if not bin in svlengths_reads:
            svlengths_reads[bin]=[(read_id,length)]
        else:
            svlengths_reads[bin].append((read_id,length))

    return pos,svlen,util.stdev(util.trim((v for k,v in read_starts.items()))),util.stdev(util.trim((v for k,v in read_svlengths.items())))

def resolve_bnd(svcall,cluster,config):
    mate_contig=util.most_common_top([lead.bnd_info.mate_contig for lead in cluster.leads])
    selected=[lead for lead in cluster.leads if lead.bnd_info.mate_contig==mate_contig]
    mate_ref_start=util.center([lead.bnd_info.mate_ref_start for lead in selected])
    is_first=util.most_common_top([lead.bnd_info.is_first for lead in selected])
    is_reverse=util.most_common_top([lead.bnd_info.is_reverse for lead in selected])
    svcall.alt=(("N" if is_first else "") +
                ("]" if is_reverse else "[" ) +
                f"{mate_contig}:{mate_ref_start}" +
                ("]" if is_reverse else "[" ) +
                ("N" if not is_first else ""))
    svcall.support=len(set(k.read_qname for k in selected))
    cluster.leads=selected
    svcall.bnd_info=SVCallBNDInfo(mate_contig=mate_contig, mate_ref_start=mate_ref_start, is_first=is_first, is_reverse=is_reverse)
    svcall.set_info("CHR2",mate_contig)

def call_groups(svgroups,config,task):
    for group in svgroups:
        svcall=call_group(group,config,task)
        if svcall!=None:
            yield svcall

def call_group(svgroup,config,task):
    """For combining grouped SV calls from multiple .snf samples into one SV call"""
    first_cand=svgroup.candidates[0]

    #Filtering
    samples_count=float(len(config.snf_input_info))
    sample_internal_ids=set(sample["internal_id"] for sample in config.snf_input_info)
    total_count=len(svgroup.included_samples)
    pass_count=sum(cand.qc==True for cand in svgroup.candidates)
    qc=(pass_count > 0 and pass_count/samples_count >= config.combine_high_confidence) or (total_count/samples_count >= config.combine_low_confidence and total_count >= config.combine_low_confidence_abs)

    if not qc:
        return None

    if (not config.combine_output_filtered) and not any(cand.qc and cand.filter=="PASS" for cand in svgroup.candidates):
        return None

    if config.output_rnames:
        rnames=[]
    else:
        rnames=None

    genotypes={}
    genotyped_count=0
    for cand in svgroup.candidates:
        if config.output_rnames and cand.rnames!=None:
            rnames.extend(cand.rnames)
        if not 0 in cand.genotypes:
            cand.genotypes[0]=(".",".",0,0,cand.support)
        if cand.sample_internal_id in genotypes:
            #Intra-sample merging
            a,b,gt_qual,dr,dv=cand.genotypes[0]
            curr_a,curr_b,curr_gt_qual,curr_dr,curr_dv,curr_id=genotypes[cand.sample_internal_id]
            new_id=curr_id+","+config.id_prefix+cand.id
            if (curr_a==".") or (a != "." and (a,b) >= (curr_a,curr_b)):
                genotypes[cand.sample_internal_id]=(a,b,gt_qual,dr,dv,new_id)
            else:
                genotypes[cand.sample_internal_id]=(curr_a,curr_b,curr_gt_qual,curr_dr,curr_dv,new_id)
        else:
            a,b,gt_qual,dr,dv=cand.genotypes[0]
            genotypes[cand.sample_internal_id]=(a,b,gt_qual,dr,dv,config.id_prefix+cand.id)
        genotyped_count+=1

    for sample_internal_id in sample_internal_ids:
        if sample_internal_id in genotypes:
            continue
        coverage=svgroup.coverages_nonincluded[sample_internal_id]
        if coverage >= config.combine_null_min_coverage:
            genotypes[sample_internal_id]=(0,0,0,coverage,0,"NULL")
        else:
            genotypes[sample_internal_id]=(".",".",0,coverage,0,"NULL")

    if config.combine_consensus:
        genotypes_consensus={}
        for a,b,gt_qual,dr,dv in genotypes.values():
            if not (a,b) in genotypes_consensus:
                genotypes_consensus[(a,b)]={"count":0,"qual":list(),"dr":list(),"dv":list()}
            genotypes_consensus[(a,b)]["count"]+=1
            genotypes_consensus[(a,b)]["qual"].append(gt_qual)
            genotypes_consensus[(a,b)]["dr"].append(dr)
            genotypes_consensus[(a,b)]["dv"].append(dv)
        most_common_count=genotypes_consensus[sorted(genotypes_consensus,key=lambda k: genotypes_consensus[k]["count"],reverse=True)[0]]["count"]
        most_common_gt=[gt for gt in genotypes_consensus if genotypes_consensus[gt]["count"]==most_common_count]
        cons_a,cons_b=max(most_common_gt)
        consensus_info=genotypes_consensus[(cons_a,cons_b)]
        genotypes={0:(cons_a,cons_b,int(sum(consensus_info["qual"])/consensus_info["count"]),sum(consensus_info["dr"]),sum(consensus_info["dv"]))}
        if cons_a!=1 and cons_b!=1:
            return None

    svcall_pos=int(util.median(cand.pos for cand in svgroup.candidates))
    svcall_svlen=int(util.median(cand.svlen for cand in svgroup.candidates))
    svcall_end=svcall_pos+abs(svcall_svlen)
    svcall_alt=first_cand.alt
    svcall_alt_mindist=abs(len(svcall_alt)-svcall_svlen)
    if first_cand.svtype=="INS":
        for cand in svgroup.candidates:
            dist=abs(len(cand.alt)-svcall_svlen)
            if dist < svcall_alt_mindist:
                svcall_alt_mindist=dist
                svcall_alt=cand.alt

    svcall=SVCall(contig=first_cand.contig,
                  pos=svcall_pos,
                  id=f"{first_cand.svtype}.{task.sv_id:X}M{task.id:X}",
                  ref="N",
                  alt=svcall_alt,
                  qual=round(util.mean(int(cand.qual) for cand in svgroup.candidates)),
                  filter="PASS",
                  info=dict(),
                  svtype=first_cand.svtype,
                  svlen=svcall_svlen,
                  end=svcall_end,
                  genotypes=genotypes,
                  precise=sum(int(cand.precise) for cand in svgroup.candidates)/float(len(svgroup.candidates)) > 0.5,
                  support=round(util.mean(cand.support for cand in svgroup.candidates)),
                  rnames=rnames,
                  postprocess=None,
                  qc=True,
                  nm=-1,
                  fwd=sum(cand.fwd for cand in svgroup.candidates),
                  rev=sum(cand.rev for cand in svgroup.candidates),
                  coverage_upstream=util.mean_or_none_round(cand.coverage_upstream for cand in svgroup.candidates if cand.coverage_upstream!=None),
                  coverage_start=util.mean_or_none_round(cand.coverage_start for cand in svgroup.candidates if cand.coverage_start!=None),
                  coverage_center=util.mean_or_none_round(cand.coverage_center for cand in svgroup.candidates if cand.coverage_center!=None),
                  coverage_end=util.mean_or_none_round(cand.coverage_end for cand in svgroup.candidates if cand.coverage_end!=None),
                  coverage_downstream=util.mean_or_none_round(cand.coverage_downstream for cand in svgroup.candidates if cand.coverage_downstream!=None) )

    svcall.set_info("STDEV_POS",util.stdev(cand.pos for cand in svgroup.candidates))
    svcall.set_info("STDEV_LEN",util.stdev(cand.svlen for cand in svgroup.candidates))

    if abs(svcall.svlen) < config.minsvlen:
        return None

    task.sv_id+=1

    return svcall

def classify_splits(read,leads,config,main_contig):
    minsvlen_screen=config.minsvlen_screen
    maxsvlen_other=minsvlen_screen*5

    leads.sort(key=lambda ld: ld.qry_start)
    last=leads[0]
    last.svtypes_starts_lens=[]

    if last.qry_start >= config.long_ins_length*0.5:
        last.svtypes_starts_lens.append(("INS",last.ref_start,None))

    for i in range(1,len(leads)):
        curr=leads[i]
        curr.svtypes_starts_lens=[]
        qry_dist_abs=abs(curr.qry_start - last.qry_end)

        if curr.contig == last.contig:
            rev=(curr.strand == "-")
            fwd=not rev
            if curr.strand == last.strand:
                #
                #INS, DEL, DUP
                #
                if (fwd and (curr.qry_start - last.qry_end) >= minsvlen_screen
                    and (curr.ref_start - last.ref_end) < maxsvlen_other
                    and (curr.qry_start - last.qry_end) - (curr.ref_start - last.ref_end) >= minsvlen_screen):
                    #INS, FWD
                    svstart=curr.ref_start
                    svlen=(curr.qry_start - last.qry_end)
                    if svlen <= config.dev_seq_cache_maxlen:
                        curr.seq=read.query_sequence[last.qry_end:curr.qry_start]
                    else:
                        curr.seq=None
                    curr.svtypes_starts_lens.append(("INS",svstart,svlen))

                elif (rev and (curr.qry_start - last.qry_end) >= minsvlen_screen
                      and (last.ref_start - curr.ref_end) < maxsvlen_other
                      and (curr.qry_start - last.qry_end) - (last.ref_start - curr.ref_end) >= minsvlen_screen):
                    #INS, REV
                    svstart=last.ref_start
                    svlen=(curr.qry_start - last.qry_end)
                    if svlen <= config.dev_seq_cache_maxlen:
                        curr.seq=read.query_sequence[last.qry_end:curr.qry_start]
                    else:
                        curr.seq=None
                    curr.svtypes_starts_lens.append(("INS",svstart,svlen))

                elif (fwd and (curr.ref_start - last.ref_end) >= minsvlen_screen
                      and qry_dist_abs < maxsvlen_other
                      and (curr.ref_start - last.ref_end) - (curr.qry_start - last.qry_end) >= minsvlen_screen):
                        #DEL, FWD
                        svstart=curr.ref_start
                        svlen=(curr.ref_start - last.ref_end)
                        curr.svtypes_starts_lens.append(("DEL",svstart,-svlen))

                elif (rev and (last.ref_start - curr.ref_end) >= minsvlen_screen
                      and qry_dist_abs < maxsvlen_other
                      and (last.ref_start - curr.ref_end) - (curr.qry_start - last.qry_end) >= minsvlen_screen):
                        #DEL, REV
                        svstart=last.ref_start
                        svlen=(last.ref_start - curr.ref_end)
                        curr.svtypes_starts_lens.append(("DEL",svstart,-svlen))

                elif fwd and curr.ref_start <= last.ref_end:
                    if qry_dist_abs < maxsvlen_other:
                        #DUP, FWD
                        svstart=curr.ref_start
                        svlen=(last.ref_end - curr.ref_start)
                        if svlen >= minsvlen_screen:
                            curr.svtypes_starts_lens.append(("DUP",svstart,svlen))

                elif rev and last.ref_start <= curr.ref_end:
                    if qry_dist_abs < maxsvlen_other:
                        #DUP, REV
                        svstart=last.ref_start
                        svlen=(curr.ref_end - last.ref_start)
                        if svlen >= minsvlen_screen:
                            curr.svtypes_starts_lens.append(("DUP",svstart,svlen))

            elif qry_dist_abs < maxsvlen_other:
                #
                #INV
                #
                if fwd and curr.ref_start <= last.ref_start:
                    #CASE B
                    svstart=curr.ref_start
                    svlen=last.ref_start-curr.ref_start
                    if svlen >= minsvlen_screen:
                        curr.svtypes_starts_lens.append(("INV",svstart,svlen))

                elif fwd and curr.ref_start > last.ref_start:
                    #CASE C
                    svstart=last.ref_start
                    svlen=curr.ref_start-last.ref_start
                    if svlen >= minsvlen_screen:
                        curr.svtypes_starts_lens.append(("INV",svstart,svlen))

                elif rev and curr.ref_end >= last.ref_end:
                    #CASE A
                    svstart=last.ref_end
                    svlen=curr.ref_end-last.ref_end
                    if svlen >= minsvlen_screen:
                        curr.svtypes_starts_lens.append(("INV",svstart,svlen))

                elif rev and curr.ref_end < last.ref_end:
                    #CASE D
                    svstart=curr.ref_end
                    svlen=last.ref_end-curr.ref_end
                    if svlen >= minsvlen_screen:
                        curr.svtypes_starts_lens.append(("INV",svstart,svlen))

        elif qry_dist_abs < maxsvlen_other:
            #
            #BND
            #
            if curr.contig == main_contig:
                a,b=curr,last
            else:
                a,b=last,curr

            if a.contig == main_contig:
                is_first=a.qry_start < b.qry_start
                if is_first:
                    if a.strand=="+":
                        svstart=a.ref_end
                    else:
                        svstart=a.ref_start
                else:
                    if a.strand=="+":
                        svstart=a.ref_start
                    else:
                        svstart=a.ref_end
                a.svtypes_starts_lens.append(("BND",
                                            svstart,
                                            SVCallBNDInfo(b.contig,
                                                          b.ref_start,
                                                          is_first,
                                                          a.strand!=b.strand)))
        last=curr



#!/usr/bin/env python3
#
# Sniffles2
# A fast structural variant caller for long-read sequencing data
#
# Created: 27.08.2021
# Author:  Moritz Smolka
# Contact: moritz.g.smolka@gmail.com
#

from dataclasses import dataclass
import gc
import math

import pysam

from sniffles import leadprov
from sniffles import cluster
from sniffles import sv
from sniffles import postprocessing
from sniffles import snf
from sniffles import util

@dataclass
class Task:
    id: int
    sv_id: int
    contig: str
    start: int
    end: int
    assigned_process_id: int=None
    lead_provider: object=None
    bam: object=None
    tandem_repeats: list=None
    genotype_svs: list=None

    def build_leadtab(self,config):
        assert(self.lead_provider==None)

        self.bam=pysam.AlignmentFile(config.input, config.input_mode, require_index=True)
        self.lead_provider=leadprov.LeadProvider(config,self.id*config.task_read_id_offset_mult)
        externals=self.lead_provider.build_leadtab(self.contig,self.start,self.end,self.bam)
        return externals,self.lead_provider.read_count

    def call_candidates(self,keep_qc_fails,config):
        candidates=[]
        for svtype in sv.TYPES:
            for svcluster in cluster.resolve(svtype,self.lead_provider,config,self.tandem_repeats):
                for svcall in sv.call_from(svcluster,config,keep_qc_fails,self):
                    candidates.append(svcall)

        self.coverage_average_fwd,self.coverage_average_rev=postprocessing.coverage(candidates,self.lead_provider,config)
        self.coverage_average_total=self.coverage_average_fwd+self.coverage_average_rev
        return candidates

    def finalize_candidates(self,candidates,keep_qc_fails,config):
        passed=[]
        for svcall in candidates:
            svcall.qc=svcall.qc and postprocessing.qc_sv(svcall,config)
            if not keep_qc_fails and not svcall.qc:
                continue
            svcall.qc=svcall.qc and postprocessing.qc_sv_support(svcall,self.coverage_average_total,config)
            if not keep_qc_fails and not svcall.qc:
                continue

            postprocessing.annotate_sv(svcall,config)

            svcall.qc=svcall.qc and postprocessing.qc_sv_post_annotate(svcall,config)
            if not keep_qc_fails and not svcall.qc:
                continue

            svcall.finalize() #Remove internal information (not written to output) before sending to mainthread for VCF writing
            passed.append(svcall)
        return passed

    def combine(self,config):
        samples_headers_snf={}
        for snf_info in config.snf_input_info:
            snf_in=snf.SNFile(config,open(snf_info["filename"],"rb"),filename=snf_info["filename"])
            snf_in.read_header()
            samples_headers_snf[snf_info["internal_id"]]=(snf_info["filename"],snf_in.header,snf_in)

            if not config.snf_combine_keep_open:
                snf_in.close()

        svcalls=[]

        #block_groups_keep_threshold=5000
        #TODO: Parameterize
        bin_min_size=config.combine_min_size
        bin_max_candidates=max(25,int(len(config.snf_input_info)*0.5))
        overlap_abs=config.combine_overlap_abs

        sample_internal_ids=set()
        for sample_internal_id,(sample_filename,sample_header,sample_snf) in samples_headers_snf.items():
            sample_internal_ids.add(sample_internal_id)

        #
        # Load candidate SVs from all samples for each block separately and cluster them based on start position
        #
        had=False
        candidates_processed=0
        svtypes_candidates_bins={svtype: {} for svtype in sv.TYPES}
        groups_keep={svtype:list() for svtype in sv.TYPES}
        for block_index in range(self.start,self.end+config.snf_block_size,config.snf_block_size):
            samples_blocks={}
            for sample_internal_id,(sample_filename,sample_header,sample_snf) in samples_headers_snf.items():
                blocks=sample_snf.read_blocks(self.contig,block_index)
                samples_blocks[sample_internal_id]=blocks

            for svtype in sv.TYPES:
                bins={}
                #svcandidates=[]
                for sample_internal_id,(sample_filename,sample_header,sample_snf) in samples_headers_snf.items():
                    blocks=samples_blocks[sample_internal_id]
                    if blocks==None:
                        continue
                    for block in blocks:
                        for cand in block[svtype]:
                            #if config.combine_pass_only and (cand.qc==False or cand.filter!="PASS"):
                            #    continue

                            cand.sample_internal_id=sample_internal_id

                            bin=int(cand.pos/bin_min_size)*bin_min_size
                            if not bin in bins:
                                bins[bin]=[cand]
                            else:
                                bins[bin].append(cand)
                        candidates_processed+=len(block[svtype])

                if len(bins)==0:
                    continue

                curr_bin=0
                size=0
                svcands=[]
                keep=groups_keep[svtype]
                sorted_bins=sorted(bins)
                last_bin=sorted_bins[-1]
                for curr_bin in sorted_bins:
                    svcands.extend(bins[curr_bin])
                    size+=bin_min_size

                    if (not config.combine_exhaustive and len(svcands) >= bin_max_candidates) or curr_bin == last_bin:
                        if len(svcands)==0:
                            size=0
                            continue
                        prevkept=len(keep)
                        svgroups=cluster.resolve_block_groups(svtype,svcands,keep,config)
                        groups_call=[]
                        keep=[]
                        for group in svgroups:
                            coverage_bin=int(group.pos_mean/config.coverage_binsize_combine)*config.coverage_binsize_combine
                            for non_included_sample in sample_internal_ids-group.included_samples:
                                if samples_blocks[non_included_sample]!=None and coverage_bin in samples_blocks[non_included_sample][0]["_COVERAGE"]:
                                    coverage=samples_blocks[non_included_sample][0]["_COVERAGE"][coverage_bin]
                                else:
                                    coverage=0
                                if non_included_sample in group.coverages_nonincluded:
                                    group.coverages_nonincluded[non_included_sample]=max(coverage,group.coverages_nonincluded[non_included_sample])
                                else:
                                    group.coverages_nonincluded[non_included_sample]=coverage

                            if abs(group.pos_mean - curr_bin) < max(size*0.5,overlap_abs):
                                keep.append(group)
                            else:
                                groups_call.append(group)
                        svcalls.extend(sv.call_groups(groups_call,config,self))
                        size=0
                        svcands=[]

                groups_keep[svtype]=keep

        for svtype in groups_keep:
            svcalls.extend(sv.call_groups(groups_keep[svtype],config,self))

        return svcalls,candidates_processed

@dataclass
class Process:
    id: int
    process: object=None
    pipe_main: object=None
    externals: list=None

def Main(proc_id,config,pipe):
    try:
        if config.dev_profile:
            import cProfile
            cProfile.runctx("Main_Internal(proc_id,config,pipe)",globals(),locals(),sort="cumulative")
        else:
            Main_Internal(proc_id,config,pipe)
    except Exception as e:
        pipe.send(["worker_exception",""])
        raise e

def Main_Internal(proc_id,config,pipe):
    tasks={}
    while True:
        command,arg=pipe.recv()

        if command=="call_sample":
            task=arg
            result={}

            if config.snf != None or config.no_qc:
                qc=False
            else:
                qc=True

            _,read_count=task.build_leadtab(config)
            svcandidates=task.call_candidates(qc,config)
            svcalls=task.finalize_candidates(svcandidates,not qc,config)
            if config.no_qc:
                result["svcalls"]=svcalls
            else:
                result["svcalls"]=[s for s in svcalls if s.qc]

            if config.snf != None: # and len(svcandidates):
                snf_filename=f"{config.snf}.tmp_{task.id}.snf"

                with open(snf_filename,"wb") as handle:
                    snf_out=snf.SNFile(config,handle)
                    for cand in svcandidates:
                        snf_out.store(cand)
                    snf_out.annotate_block_coverages(task.lead_provider)
                    snf_out.write_and_index()
                    handle.close()
                result["snf_filename"]=snf_filename
                result["snf_index"]=snf_out.get_index()
                result["snf_total_length"]=snf_out.get_total_length()
                result["snf_candidate_count"]=len(svcandidates)
                result["has_snf"]=True
            else:
                result["has_snf"]=False

            #if config.vcf != None:
            #    svcalls=task.finalize_candidates(svcandidates,config)
            #    result["svcalls"]=svcalls

            result["task_id"]=task.id
            result["processed_read_count"]=read_count
            result["coverage_average_total"]=task.coverage_average_total
            pipe.send(["return_call_sample",result])
            del task
            gc.collect()

        elif command=="genotype_vcf":
            task=arg
            result={}

            qc=False
            _,read_count=task.build_leadtab(config)
            svcandidates=task.call_candidates(qc,config=config)
            svcalls=task.finalize_candidates(svcandidates,not qc,config=config)

            binsize=5000
            binedge=int(binsize/10)
            genotype_svs_svtypes_bins={svtype:{} for svtype in sv.TYPES}
            for genotype_sv in task.genotype_svs:
                genotype_sv.genotype_match_sv=None
                genotype_sv.genotype_match_dist=math.inf

                if not genotype_sv.svtype in genotype_svs_svtypes_bins:
                    #TODO: Warn about unsupported SVTYPE
                    continue

                bins=[int(genotype_sv.pos/binsize)*binsize]
                if genotype_sv.pos%binsize < binedge:
                    bins.append((int(genotype_sv.pos/binsize)-1)*binsize)
                if genotype_sv.pos%binsize > binsize-binedge:
                    bins.append((int(genotype_sv.pos/binsize)+1)*binsize)

                for bin in bins:
                    if not bin in genotype_svs_svtypes_bins[genotype_sv.svtype]:
                         genotype_svs_svtypes_bins[genotype_sv.svtype][bin]=[]
                    genotype_svs_svtypes_bins[genotype_sv.svtype][bin].append(genotype_sv)

            for cand in svcandidates:
                bin=int(cand.pos/binsize)*binsize
                if not bin in genotype_svs_svtypes_bins[cand.svtype]:
                    continue
                if cand.svtype=="BND":
                    for genotype_sv in genotype_svs_svtypes_bins[cand.svtype][bin]:
                        dist=abs(genotype_sv.pos - cand.pos)
                        #if minlen>0 and dist < genotype_sv.genotype_match_dist and dist <= config.cluster_merge_bnd * 2:
                        if dist < genotype_sv.genotype_match_dist and dist <= config.cluster_merge_bnd:
                            if cand.bnd_info.mate_contig==genotype_sv.bnd_info.mate_contig:
                                genotype_sv.genotype_match_sv=cand
                                genotype_sv.genotype_match_dist=dist
                else:
                    for genotype_sv in genotype_svs_svtypes_bins[cand.svtype][bin]:
                        dist=abs(genotype_sv.pos - cand.pos) + abs(abs(genotype_sv.svlen) - abs(cand.svlen))
                        minlen=float(min(abs(genotype_sv.svlen),abs(cand.svlen)))
                        if minlen>0 and dist < genotype_sv.genotype_match_dist and dist <= config.combine_match * math.sqrt(minlen) and dists <= config.combine_match_max:
                            genotype_sv.genotype_match_sv=cand
                            genotype_sv.genotype_match_dist=dist

            postprocessing.coverage(task.genotype_svs,task.lead_provider,config)

            #Determine genotypes for unmatched input SVs
            for svcall in task.genotype_svs:
                coverage_list=[svcall.coverage_start,svcall.coverage_center,svcall.coverage_end]
                coverage_list=[c for c in coverage_list if c!=None]
                if len(coverage_list)==0:
                    return
                coverage=round(sum(coverage_list)/len(coverage_list))
                svcall.genotypes={}
                if coverage>0:
                    svcall.genotypes[0]=(0,0,0,coverage,0)
                else:
                    svcall.genotypes[0]=config.genotype_none

            result={}
            result["task_id"]=task.id
            result["processed_read_count"]=read_count
            result["svcalls"]=task.genotype_svs
            pipe.send(["return_genotype_vcf",result])
            del task
            gc.collect()

        elif command=="combine":
            task=arg
            result={}

            result["svcalls"],candidates_processed=task.combine(config)

            result["task_id"]=task.id
            result["processed_read_count"]=candidates_processed
            pipe.send(["return_combine",result])
            del task
            gc.collect()

        elif command=="finalize":
            return
